/*
 * sgsterm.c - attempt at a gnome terminal program.
 *	NOT another xterm, but an RS232 serial terminal emulator...
 */

#include <gnome.h>

#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <tty.h>
#include <zvt/zvtterm.h>
#define FONT "-misc-fixed-medium-r-normal--14-*-*-*-*-*-iso8859-1"

int g_verbose;
int g_serfd = -1;	/* serial tty file descriptor */
FILE *g_asendfp;	/* NULL or curent file for ascii send */
char *g_asend_fname;
char g_asend_prompt = '>';
char *g_logfile_fname;
FILE *g_logfp;		/* NULL or file used for logging */
int g_hexmode = 0;	/* if 1, display incomming data in hex */
ZvtTerm *g_term;
int g_baud = 9600;
char *g_portname = "/dev/ttyS11";
GnomeAppBar *g_appbar;

GtkWidget *create_file_selector(char *name, 
				GtkSignalFunc func, char *default_fname);
static void
format_as_hex(ZvtTerm *term, char *buffer, int count, FILE *fp);

/* this is usually defined by autoconf but we're just using simple makefiles */
#define VERSION "1.1"

/* table of sgsterm-specific command-line options */
struct poptOption sgsterm_optionTable[] = {
	{ "bps", 'b', POPT_ARG_INT, &g_baud, 0,
	  "signaling rate in bits-per-second", "BPS" },
	{ "line", 'l', POPT_ARG_STRING, &g_portname, 0,
	  "serial line device name", "DEV" },

	{ NULL, 0, 0, NULL, 0 }
};


/*
 * print character, with control chars in ^ style
 */
void c_putc(char c, FILE *fp)
{
	switch(c & 0xff) {
	case '\n':
		fprintf(fp, "^J\n");
		break;
	case '\r':
		fprintf(fp, "^M\r");
		break;
	default:
		if(c < ' ') {
			fprintf(fp, "^%c", c + '@');
		} else {
			fputc(c, fp);
		}
	}
}

void
c_puts(FILE *fp, char *buf, int n)
{
	int i;
	for(i = 0; i < n; i++) {
		c_putc(buf[i], fp);
	}
	fflush(fp);
}

/*
 * Open serial port and set up for comunication
 */
int ser_open(char *dev, int baudrate)
{
	int tty_fd;

	tty_fd = tty_open(dev);
	if(tty_fd < 0) {
		return -1;
	}

	if(tty_set(tty_fd, baudrate, 8, 1, NO_PARITY) < 0) {
		tty_close(tty_fd);
		return -1;
	}

	if(tty_raw(tty_fd, 0) < 0) {
		tty_close(tty_fd);
		return -1;
	}
	return tty_fd;
}

/* continue ascii send: if prompt character has been recieved, send next line.
 * TODO: prefs dialog box for setting prompt character and line/char delays
 * 	option to use delays only instead of prompts
 *	multi-character prompt, possibly regular expression.
 */
void
do_asend(char *rbuf, int rcount, FILE *fp)
{
	char buf[1024];
	char *cp;

	if(memchr(rbuf, g_asend_prompt, rcount)) {
		if(fgets(buf, 1024, fp) != NULL) {
			if(cp = strchr(buf, '\n'))
				*cp = 0;
			write(g_serfd, buf, strlen(buf));
			write(g_serfd, "\r", 1);
		} else {
			gnome_appbar_set_status(g_appbar, "ascii send completed");
			fclose(fp);
			g_asendfp = NULL;
		}
	}
}

/* callback function when data is available on the serial port 
 * We should probably copy in most of zvt_term_readdata
 * instead of using zvt_term_feed()
 */
static void
serial_handler(gpointer data, gint fd, GdkInputCondition condition)
{
	int count;
	char buffer[4096];
	int saveerrno;

	ZvtTerm *term;
	term = (ZvtTerm *) data;

	saveerrno = EAGAIN;
	while((saveerrno == EAGAIN) && (count = read (fd, buffer, 4096)) > 0) {
		saveerrno = errno;

		if(g_hexmode) {
			format_as_hex(term, buffer, count, g_logfp);
		} else {
			zvt_term_feed(term, buffer, count);
			if(g_logfp)
				fwrite(buffer, 1, count, g_logfp);
		}
		if(g_asendfp)
			do_asend(buffer, count, g_asendfp);
	}
}

static char hexbuf[16];
static int hexcount;
/* format data received from serial port as hex. */
static void
format_as_hex(ZvtTerm *term, char *buffer, int count, FILE *fp)
{
	int i;
	char c;
	char pbuf[16];
	for(i = 0; i < count; i++) {
		c = buffer[i];
		sprintf(pbuf, "%02x ", c&0xff);
		zvt_term_feed(term, pbuf, 3);
		if(fp)
			fwrite(pbuf, 1, 3, fp);
		if(c >= ' ' && c < 0x7f) {
			hexbuf[hexcount++] = c;
		} else {
			hexbuf[hexcount++] = '.';
		}
		if(hexcount >= 16) {
			zvt_term_feed(term, hexbuf, hexcount);
			zvt_term_feed(term, "\r\n", 2);
			if(fp) {
				fwrite(hexbuf, 1, hexcount, fp);
				fwrite("\r\n", 1, 2, fp);
			}
			hexcount = 0;
		}
	}
}

static void
cb_ser_hexmode(GtkWidget *widget, gpointer data)
{
	g_hexmode = !g_hexmode;
	gnome_appbar_set_status(g_appbar, g_hexmode ? "hex display enabled"
				: "normal display enabled");
	/* should flush ascii/hex buffer */
	zvt_term_feed(g_term, "\r\n", 2);
	hexcount = 0;
}

static void
cb_ser_break(GtkWidget *widget, gpointer data)
{
	tcsendbreak(g_serfd, 0);
}

static void
cb_ser_baud(GtkWidget *widget, gpointer data)
{
	g_baud = (int) data;

	if(tty_set(g_serfd, g_baud, 8, 1, NO_PARITY) < 0) {
		fprintf(stderr, "change to %d b/s failed\n", g_baud);
	}
}

static void
cb_term_reset(GtkWidget *widget, gpointer data)
{
	zvt_term_reset(g_term, 1);
	hexcount = 0;
}

/* "callback" function (signal handler) which will quit the application*/
static void
exit_gsterm(GtkWidget *widget, gpointer data)
{
        gtk_main_quit ();
}

static void
cb_asend_gotfile(GtkWidget *w, gpointer data)
{
	GtkWidget *file_selector;
	char *old_fname;
	file_selector = (GtkWidget *)data;
	
	old_fname = g_asend_fname;
	g_asend_fname = g_strdup(gtk_file_selection_get_filename(
		GTK_FILE_SELECTION(file_selector)));

	g_asendfp = fopen(g_asend_fname, "r");
	if(!g_asendfp) {
		perror(g_asend_fname);
		g_asend_fname = old_fname;
		return;
	}
	if(old_fname)
		g_free(old_fname);
	write(g_serfd, "\r", 1);
	
}

/* prompt for filename and commence ascii download */
static void
cb_ser_asend(GtkWidget *widget, gpointer data)
{
	GtkWidget *file_selector;
	if(g_asendfp) {
		gnome_appbar_set_status(g_appbar, "ascii send canceled");
		fclose(g_asendfp);
		g_asendfp = NULL;
	} else {
		/* Create the selector */
		file_selector = create_file_selector (
			"Select file for ascii send", 
			GTK_SIGNAL_FUNC(cb_asend_gotfile), g_asend_fname);
	}
}

/* callback when logfile-selection dialog box is finished:
 * try to open the file and start logging */
static void
cb_log_gotfile(GtkWidget *w, gpointer data)
{
	GtkWidget *file_selector;
	char *old_fname;
	file_selector = (GtkWidget *)data;
	
	old_fname = g_logfile_fname;
	g_logfile_fname = g_strdup(gtk_file_selection_get_filename (
					   GTK_FILE_SELECTION(file_selector)));

	g_logfp = fopen(g_logfile_fname, "a");
	if(!g_logfp) {
		perror(g_logfile_fname);
		g_logfile_fname = old_fname;
		return;
	}
	gnome_appbar_set_status(g_appbar, "logging started");
	
	if(old_fname)
		g_free(old_fname);
}

/* prompt for filename to begin logging to file.
 * Or cancel logging if already enabled.
 */
static void
cb_ser_logfile(GtkWidget *widget, gpointer data)
{
	GtkWidget *file_selector;
	if(g_logfp) {
		gnome_appbar_set_status(g_appbar, "logfile closed");
		fclose(g_logfp);
		g_logfp = NULL;
	} else {
		/* Create the selector */
		file_selector = create_file_selector (
			"Select file for logging", 
			GTK_SIGNAL_FUNC(cb_log_gotfile), g_logfile_fname);
	}
}

/* build and show a file selector */
GtkWidget *create_file_selector(char *name, 
				GtkSignalFunc func, char *default_fname)
{
	GtkWidget *file_selector;

     	file_selector = gtk_file_selection_new(name);
     	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->ok_button),
			    "clicked", func, file_selector);
                             
	/* Ensure that the dialog box is destroyed when the user clicks a button. */
     
	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->ok_button),
				   "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
				   (gpointer) file_selector);

	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->cancel_button),
				   "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
				   (gpointer) file_selector);

	if(default_fname) {
		gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector), default_fname);
	}
     
	gtk_widget_show (file_selector);

}

/* callback function for when the window closes */
static int
delete_event(GtkWidget *widget, gpointer data)
{
        gtk_main_quit ();
        return FALSE; /* false means continue with closing the window */
}


/* a callback for the about menu item, it will display a simple "About"
 * dialog box standard to all gnome applications
 */
static void
about_gsterm(GtkWidget *widget, gpointer data)
{
        GtkWidget *box;
        const char *authors[] = {
		"Steve Tell",
                NULL
        };

        box = gnome_about_new(/*title: */ "Sgsterm",
                              /*version: */VERSION,
                              /*copyright: */ "(C) 2001 Steve Tell.",
                              /*authors: */authors,
                              /*other comments: */
                              "A simple serial terminal emulator.",
                              NULL);
        gtk_widget_show(box);
}

/* define the menus here */

static GnomeUIInfo file_menu [] = {
        /* some item which is not one of the standard ones, the null
         * would be the callback, however we don't want to really do anything */
        GNOMEUIINFO_ITEM_NONE("Ascii send..", "send an ascii file",
			      cb_ser_asend),
        GNOMEUIINFO_ITEM_NONE("Log to file", "log incomming data to a file",
			      cb_ser_logfile),
        GNOMEUIINFO_ITEM_NONE("Reset", "Reset terminal and clear screen",
			      cb_term_reset),

        GNOMEUIINFO_ITEM_NONE("Send Break", "send Break out serial port",
			      cb_ser_break),
        /* standard exit item */
        GNOMEUIINFO_MENU_EXIT_ITEM(exit_gsterm,NULL),
        GNOMEUIINFO_END
};

static GnomeUIInfo baud_menu [] = {
        GNOMEUIINFO_ITEM_DATA("19200","set serial bit rate to 19200 b/s", cb_ser_baud, (gpointer)19200, NULL),
        GNOMEUIINFO_ITEM_DATA("9600","set serial bit rate to 9600 b/s", cb_ser_baud, (gpointer)9600, NULL),
        GNOMEUIINFO_ITEM_DATA("1200","set serial bit rate to 1200 b/s", cb_ser_baud, (gpointer)1200, NULL),
        GNOMEUIINFO_END
};

static GnomeUIInfo settings_menu [] = {
        GNOMEUIINFO_SUBTREE("Bit Rate", baud_menu),
        GNOMEUIINFO_ITEM_NONE("Hex Mode", "Display incomming data in hex",
			      cb_ser_hexmode),
        GNOMEUIINFO_END
};

static GnomeUIInfo help_menu [] = {
        /* load the helpfiles for this application if available */
        GNOMEUIINFO_HELP("sgsterm"),
        /* the standard about item */
        GNOMEUIINFO_MENU_ABOUT_ITEM(about_gsterm,NULL),
        GNOMEUIINFO_END
};

/* define the main menubar */
static GnomeUIInfo main_menu [] = {
        GNOMEUIINFO_MENU_FILE_TREE(file_menu),
        GNOMEUIINFO_MENU_SETTINGS_TREE(settings_menu),
        GNOMEUIINFO_MENU_HELP_TREE(help_menu),
        GNOMEUIINFO_END
};

/* Our main function */
int
main(int argc, char *argv[])
{
        GtkWidget *app; /* pointer to our main window */
        GtkWidget *w; /* pointer to some widget */
	int input_id;
	char *fontname=FONT;

	gnomelib_register_popt_table(sgsterm_optionTable, "Sgsterm Options");
        /* initialize gnome */
        gnome_init("sgsterm", VERSION, argc, argv);

        /* create main window */
        app = gnome_app_new("sgsterm", "Gnome Serial Terminal");
        /* connect "delete_event" (happens when the window is closed */
        gtk_signal_connect(GTK_OBJECT(app), "delete_event",
                           GTK_SIGNAL_FUNC(delete_event),
                           NULL);

        /* add the menus to the main window */
        gnome_app_create_menus(GNOME_APP(app), main_menu);

        /* setup appbar (bottom of window bar for status, menu hints and
         * progress display) */
	g_appbar = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_USER);
        gnome_app_set_statusbar(GNOME_APP(app), g_appbar);

        /* make menu hints display on the appbar */
        gnome_app_install_menu_hints(GNOME_APP(app), main_menu);

        /* set up window contents */
        g_term = (ZvtTerm *)zvt_term_new_with_size(80, 24);
	zvt_term_set_font_name(ZVT_TERM (g_term), fontname);
	zvt_term_set_blink (ZVT_TERM (g_term), TRUE);
	zvt_term_set_bell (ZVT_TERM (g_term), TRUE);
	zvt_term_set_scrollback(ZVT_TERM (g_term), 100);
	zvt_term_set_scroll_on_keystroke (ZVT_TERM (g_term), TRUE);
	zvt_term_set_scroll_on_output (ZVT_TERM (g_term), FALSE);
	zvt_term_set_background (ZVT_TERM (g_term), NULL, 0, 0);
	zvt_term_set_wordclass (ZVT_TERM (g_term), "-A-Za-z0-9/_:.,?+%=");

        /* add those contents to the window */
        gnome_app_set_contents(GNOME_APP(app), GTK_WIDGET(g_term));

	/* Open serial port */
	g_serfd = ser_open(g_portname, g_baud);
	if(g_serfd < 0) {
		perror(g_portname);
		exit(1);
	}

        /* show all the widgets on the main window and the window itself */
        gtk_widget_show_all(app);

	/*zvt sends ascii resulting from keystrokes directly to serial port. */
	g_term->vx->vt.keyfd = g_serfd;

	/* serial characters go to zvt via our callback */
	input_id = gdk_input_add(g_serfd, GDK_INPUT_READ, 
				 serial_handler, g_term);

        /* run the main loop */
        gtk_main();
	tty_close(g_serfd);
        return 0;
}
