/*
 * General-purpose serial I/O routines.  Oriented towards talking to
 * modems and embedded devices.   Not perfect, but still quite useful.
 * by Steve Tell

 * Copyright (C) 1998 Stephen G. Tell, tell@cs.unc.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 *
 */

#include <config.h>
#ifdef HAVE_POSIX_TERMIOS
#include <termios.h>
#else
#error "sorry, we only support posix termios for terminal control"
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include "tty.h"

extern int g_verbose;

typedef struct _tty_speed
{
	int code;
	int speed;
} tty_speed;

tty_speed tty_speeds[] =
{
	{B50, 50},
	{B75, 75},
	{B110, 110},
	{B134, 134},
	{B150, 150},
	{B200, 200},
	{B300, 300},
	{B600, 600},
	{B1200, 1200},
	{B1800, 1800},
	{B2400, 2400},
	{B4800, 4800},
	{B9600, 9600},
	{B19200, 19200},
	{B38400, 38400},
#ifdef B57600
	{B57600, 57600},
#endif
#ifdef B115200
	{B115200, 115200},
#endif
#ifdef B230400
	{B230400, 230400},
#endif
	{-1, -1}
};


typedef struct _tty_size
{
	int code;
	int size;
	
} tty_size;

tty_size tty_sizes[] =
{
	{CS5, 5},
	{CS6, 6},
	{CS7, 7},
	{CS8, 8},
	{-1, -1}
};

/*
 * Open serial terminal device.  
 * Will not block opening the device, but subsequently arranges for reads
 * to block.
 */
int tty_open(char *device)
{
	int fd;
	int flags;
	
	/* set O_NDELAY so open() returns without waiting for carrier */
	if((fd = open(device, O_RDWR | O_NDELAY | O_NOCTTY, 0)) < 0)
	{
		perror(device);
		return -1;
	}								   
	
	flags = fcntl(fd, F_GETFL, 0);
	if(g_verbose)
		fprintf(stderr, "fcntl flags=0x%x\n", flags);
	if (flags < 0) {
		fprintf(stderr, "fcntl(%s, F_GETFL): %s", device, strerror(errno));
		close(fd);
		return -1;

	}
	if(fcntl(fd, F_SETFL, flags & ~O_NDELAY) < 0) {
		fprintf(stderr, "fcntl(%s, F_SETFL): %s", device, strerror(errno));
		close(fd);
		return -1;
	}
	return fd;
}

/* we define a corresponding close routine just for symetry's sake. */
void tty_close(int fd)
{
	close(fd);
}

int tty_isspeed(int speed)
{
	tty_speed *s;

	for (s = &tty_speeds[0]; s->code != -1; s++) {
		if (s->speed == speed) {
			return s->code;
		}
	}
	return -1;
}

/*
 * set basic serial parameters:
 *	bit rate, number of data bits, number of stop bits, parity.
 */
int tty_set(int fd, int speed, int data, int stop, int parity)
{
	int scode;
	tty_size *d;
	struct termios t;

	if (tcgetattr(fd, &t) < 0) {
		perror("tcgetattr");
		return -1;
	}
	cfmakeraw(&t);
	t.c_cc[VMIN]  = 1;
	t.c_cc[VTIME] = 0;


	scode = tty_isspeed(speed);
	if (scode == -1) {
		fprintf(stderr, "speed %d not supported\n", speed);
		return -1;
	}
	cfsetispeed(&t, scode);
	cfsetospeed(&t, scode); 

	for (d = tty_sizes; d->size != data; d++) {
		if (d->code == -1) {
			fprintf(stderr, "%d data bits not supported\n", data);
			return -1;
		}
	}
	t.c_cflag |= d->code;

	switch (stop) {
	case 1:
		break;
	case 2:
		t.c_cflag |= CSTOPB;
		break;
	default:
		fprintf(stderr, "%d stop bits not supported\n", stop);
		return -1;
	}

	switch (parity)	{
	case NO_PARITY:
		break;
	case ODD_PARITY:
		t.c_cflag |= PARENB;
	case EVEN_PARITY:
		t.c_cflag |= PARODD;
		break;
	default:
		fprintf(stderr, "parity must be odd, even or none\n");
		return -1;
	}

	if (tcsetattr(fd, TCSANOW, &t) < 0) {
		perror("tcsetattr(TCSANOW)");
		return -1;
	}

	return 0;
}

/* put tty in "raw" mode
 */
int tty_raw(int fd, int timeout)
{
	struct termios t;

	if (tcgetattr(fd, &t) < 0) {
		perror("tcgetattr");
		return -1;
	}

	t.c_cflag |= CLOCAL;
	t.c_cflag |= HUPCL;
	t.c_iflag = 0;
	t.c_oflag = 0;
	t.c_lflag = 0;
	t.c_line  = 0;
	if(timeout)
		t.c_cc[VMIN]  = 0;
	else
		t.c_cc[VMIN]  = 1;
	t.c_cc[VTIME] = timeout;

	if (tcsetattr(fd, TCSAFLUSH, &t) < 0) {
		perror("tcsetattr(TCSAFLUSH)");
		return -1;
	}

	return 0;
}

int tty_discard(int fd)
{
	if (tcflush(fd, TCIOFLUSH) < 0)	{
		perror("tcflush(TCIOFLUSH)");
		return -1;
	}
	return 0;
}

